/*
** Server Processes Library Example
** servproctest.c Copyright (C) 1995, 96 Alex Belits
** This code is a part of fhttpd distribution
*/
#include <stdio.h>
#include "servproc.h"

int main(int argc,char **argv){
struct http_server *server;
struct request *req;
struct httpresponse *response;
struct httpclientsession *session;
int i,j,human=1;
char tmpbuffer[10240];
char *methodnames[]={"UNKNOWN","GET","HEAD","POST","PUT"};

if(argc>1){
 if(!strcmp(argv[1],"-d")){
  human=0;
  argc--;
  argv[1]=argv[0];
  argv++;
  }
 }

server=createserver();
if(!server) return -1;

switch(servproc_init(server,human,argc,argv)){
 case 0: break;
 case APP_ERR_HUMAN:
  fprintf(stderr,"Usage: %s [-d] [appname] [username] [hostname] [portname]\n",
          server->app_progname);
  exit(1);
  break;
 case APP_ERR_CONFIG:
  fprintf(stderr,"%s: configuration error\n",server->app_progname);
  exit(1);
  break;
 case APP_ERR_READ:
  fprintf(stderr,"%s: read error\n",server->app_progname);
  exit(1);
  break;
 case APP_ERR_HOSTNAME:
  fprintf(stderr,"%s: can't resolve server hostname\n",server->app_progname);
  exit(1);
  break;
 case APP_ERR_SOCKET:
  fprintf(stderr,"%s: can't create socket\n",server->app_progname);
  exit(1);
  break;
 case APP_ERR_CONNECT:
  fprintf(stderr,"%s: can't connect\n",server->app_progname);
  exit(1);
  break;
 case APP_ERR_APPCONNECT:
  fprintf(stderr,"%s: connect error\n",server->app_progname);
  exit(1);
  break;
 case APP_ERR_USER:
  fprintf(stderr,"%s: login error\n",server->app_progname);
  exit(1);
  break;
 case APP_ERR_PASSWORD:
  fprintf(stderr,"%s: login error\n",server->app_progname);
  exit(1);
  break;
 case APP_ERR_APPLICATION:
  fprintf(stderr,"%s: application rejected by server\n",server->app_progname);
  exit(1);
  break;
 case APP_ERR_INSANE:
 case APP_ERR_DAEMON:
 case APP_ERR_AUTH:
 default:
  if(server->infd<0) exit(1);
 }

setcapabilities(server,APP_CAP_KEEPALIVE);
while((req=getrequest(server))){
 if(req->reqtype==FHTTPD_REQUEST){
  session=findsession(req,1);
  if(session){
   if(server->session_created){
    selectinstancesession(session);
    if(!strstr(req->user_agent,"Mozilla")){ 
      /* Probably it should do something better, but for testing it's enough */
     redirecttosession(session);
     deleterequest(req);
     req=NULL;
     }
    }
   }
  if(req){
   response=createresponse(1024,req->id,req->fd,1);
   if(response){
    putlinetoresponse(response,
"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\r\n"
"<HTML><HEAD><TITLE>Server process</TITLE></HEAD>\r\n"
"<BODY><H1>Server process</H1>\r\n"
"<A NAME=\"label1\">&lt;--- 1 ---</A><P>\r\n"
"<A NAME=\"label2\">&lt;--- 2 ---</A><P>\r\n");
    if(server->session_expired){
     putlinetoresponse(response,"<H3>Previous session expired</H3>\r\n");
     }
    if(server->session_created){
     putlinetoresponse(response,"<H3>New session created</H3>\r\n");
     }
    putlinetoresponse(response,
"<PRE>\r\n"
"Request: \""
);
    putmlinetoresponse(response,req->rl);
    putlinetoresponse(response,"\"\r\n");
    for(i=0;i<req->nlines;i++){
     putlinetoresponse(response,"\"");
     for(j=0;j<req->lines[i].paramc;j++){
      putmlinetoresponse(response,req->lines[i].params[j]);
      if(j<req->lines[i].paramc-1) putlinetoresponse(response,"\",\"");
      }
     putlinetoresponse(response,"\"\r\n");
     }
    putlinetoresponse(response,"</PRE>\r\n");
    sprintf(tmpbuffer,
"<TABLE BORDER=3><TR><TH>Name</TH>"
"<TH>Value</TH></TR>\r\n"
"<TR><TH>Accept</TH><TD>%s</TD></TR>\r\n"
"<TR><TH>User agent</TH><TD>%s</TD></TR>\r\n"
"<TR><TH>Content type</TH><TD>%s</TD></TR>\r\n"
"<TR><TH>Server software</TH><TD>%s</TD></TR>\r\n",
     req->accept,req->user_agent,req->content_type,req->server_software);
    putlinetoresponse(response,tmpbuffer);
    if(req->server_name){
     sprintf(tmpbuffer,"<TR><TH>Server hostname</TH><TD>%s</TD></TR>\r\n",
      req->server_name);
     putlinetoresponse(response,tmpbuffer);
     }
    if(req->server_port){
      sprintf(tmpbuffer,"<TR><TH>Server port</TH><TD>%d</TD></TR>\r\n",
        req->server_port);
     putlinetoresponse(response,tmpbuffer);
     }
    if(req->script_name){
     sprintf(tmpbuffer,"<TR><TH>Script name</TH><TD>%s</TD></TR>\r\n",
      req->script_name);
     putlinetoresponse(response,tmpbuffer);
     }
    if(req->query_string){
     sprintf(tmpbuffer,"<TR><TH>Query string</TH><TD>%s</TD></TR>\r\n",
      req->query_string);
     putlinetoresponse(response,tmpbuffer);
     }
    if(req->remote_addr){
     sprintf(tmpbuffer,"<TR><TH>Remote addres</TH><TD>%s</TD></TR>\r\n",
      req->remote_addr);
     putlinetoresponse(response,tmpbuffer);
     }
    if(req->remote_host){
     sprintf(tmpbuffer,"<TR><TH>Remote hostname</TH><TD>%s</TD></TR>\r\n",
      req->remote_host);
     putlinetoresponse(response,tmpbuffer);
     }
    if(req->remote_user){
     sprintf(tmpbuffer,"<TR><TH>Remote user</TH><TD>%s</TD></TR>\r\n",
      req->remote_user);
     putlinetoresponse(response,tmpbuffer);
     }
    sprintf(tmpbuffer,
"<TR><TH>HTTP version</TH><TD>%d.%d</TD></TR>\r\n"
"<TR><TH>HTTP method</TH><TD>%s</TD></TR>\r\n"
     ,req->ver_major,req->ver_minor,methodnames[req->method]);
    putlinetoresponse(response,tmpbuffer);
    if(req->narg){
     putlinetoresponse(response,"<TR><TH COLSPAN=2>Arguments:</TH></TR>\r\n");
     for(i=0;i<req->narg;i++){
      if(req->arg[i].params){
       if(req->arg[i].paramc==1){
        putlinetoresponse(response,"<TR><TD COLSPAN=2>");
        putmlinetoresponse(response,req->arg[i].params[0]);
        putlinetoresponse(response,"</TD></TR>\r\n");
        }else{
        if(req->arg[i].paramc==2){
         putlinetoresponse(response,"<TR><TD>");
         putmlinetoresponse(response,req->arg[i].params[0]);
         putlinetoresponse(response,"</TD><TD>");
         putmlinetoresponse(response,req->arg[i].params[1]);
         putlinetoresponse(response,"</TD></TR>\r\n");
         }
        }
       }
      }
     }
    if(req->nform){
     putlinetoresponse(response,"<TR><TH COLSPAN=2>Form input:</TH></TR>\r\n");
     for(i=0;i<req->nform;i++){
      if(req->form[i].params){
       if(req->form[i].paramc==1){
        putlinetoresponse(response,"<TR><TD COLSPAN=2>");
        putmlinetoresponse(response,req->form[i].params[0]);
        putlinetoresponse(response,"</TD></TR>\r\n");
        }else{
        if(req->form[i].paramc==2){
         putlinetoresponse(response,"<TR><TD>");
         putmlinetoresponse(response,req->form[i].params[0]);
         putlinetoresponse(response,"</TD><TD>");
         putmlinetoresponse(response,req->form[i].params[1]);
         putlinetoresponse(response,"</TD></TR>\r\n");
         }
        }
       }
      }
     }
    putlinetoresponse(response,"</TABLE>\r\n");
    putlinetoresponse(response,"<P><A HREF=\"");
    putmyurltoresponse(response,req,session,NULL);
    putlinetoresponse(response,"\">[No args]</A>");
    putlinetoresponse(response,"<A HREF=\"");
    putmyurltoresponse(response,req,session,"arg1=1");
    putlinetoresponse(response,"\">[Args 1]</A>");
    putlinetoresponse(response,"<A HREF=\"");
    putmyurltoresponse(response,req,session,"arg1=2");
    putlinetoresponse(response,"\">[Args 2]</A>");
    putlinetoresponse(response,"[Remote 2]</A>\r\n</BODY></HTML>\r\n");

    makesessionheader(response,200,NULL,NULL,session,req->keepalive);
    putlinetoheader(response,
"Pragma: no-cache\r\n"
"Cache-Control: no_cache\r\n");
    sendresponse(server,response);
    finishresponse(server,response);
    deleteresponse(response);
    }
   }
  }else{
  }
 deleterequest(req);
 session=findsessiontoexpire(server,10000,24*60*60,60);
 if(session){
  deselectinstancesession(session);
  deletesession(session);
  }
 }
return 1;
}
